import abc

from queue import deque
from typing import Callable, Coroutine, List, Optional, Union

from pymaid.core import create_task, current_task, Event, Task
from pymaid.core import get_running_loop, iscoroutine, iscoroutinefunction
from pymaid.error import BaseEx
from pymaid.ext.pools.worker import AioPool
from pymaid.utils.logger import logger_wrapper


class Handler(abc.ABC):
    '''Handle the *received* tasks.'''

    def __init__(
        self,
        *,
        on_close: Optional[List[Callable[['Handler'], None]]] = None,
        error_handler: Optional[Callable[[Task], Coroutine]] = None,
        close_on_exception: bool = False,
    ):
        self.task = None
        self.on_close = on_close or []
        self.close_on_exception = close_on_exception

        if error_handler:
            if not iscoroutinefunction(error_handler):
                raise ValueError('required error_handler as coroutinefunction')
            self.error_handler = error_handler
        else:
            self.error_handler = self.handle_error

        self.pending_tasks = deque()
        self.new_task_received = Event()
        self.closed_event = Event()
        self.is_closing = False
        self.is_closed = False

        self.task = create_task(self.run())

    @abc.abstractmethod
    async def run(self):
        raise NotImplementedError('run')

    def shutdown(self, reason: Union[None, str, Exception] = None):
        if self.is_closing:
            return
        self.logger.debug(f'{self!r} shutdown with reason={reason!r}')
        self.is_closing = True
        self.pending_tasks.append(None)
        self.new_task_received.set()

    async def join(self, reason: Optional[Union[str, Exception]] = None):
        if current_task() is self.task:
            raise RuntimeError('cannot join self')
        if not self.is_closing:
            raise RuntimeError('cannot join, call shutdown first')
        await self.closed_event.wait()

    def close(self, reason: Optional[Union[str, Exception]] = None):
        if self.is_closed:
            return
        self.logger.info(f'{self!r} close with reason={reason!r}')
        self.is_closed = True
        for task in self.pending_tasks:
            if iscoroutine(task):
                task.close()
        self.pending_tasks.clear()

        for cb in self.on_close:
            cb(self)
        self.closed_event.set()
        if not self.task.done():
            self.task.cancel()
        self.task = None

    def submit(self, task: Callable, *args, **kwargs):
        # self.logger.debug(f'{self!r} get task={task}')
        self.pending_tasks.append((task, args, kwargs))
        self.new_task_received.set()

    async def handle_error(self, error: Exception):
        '''Default error handler.

        Just write error logs.
        '''
        self.logger.error(f'{self!r} caught an unhandled error, {error!r}')

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_tpye, exc_value, exc_tb):
        self.shutdown('__aexit__')
        await self.join()

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'pending={len(self.pending_tasks)} '
            f'close_on_exception={self.close_on_exception}'
            f'>'
        )


@logger_wrapper(name='pymaid.SerialHandler')
class SerialHandler(Handler):
    '''Handle the *received* tasks *one by one*.

    This is usually used for stateful situation, like games, the previous task
    will mostly impact the later one.
    '''

    async def run(self):
        new_task_received = self.new_task_received
        pending_tasks = self.pending_tasks
        error_handler = self.error_handler

        running = True
        while running:
            await new_task_received.wait()
            # clear for reuse
            new_task_received.clear()
            while pending_tasks:
                task = pending_tasks.popleft()
                if not task:
                    running = False
                    break

                task, args, kwargs = task
                try:
                    if iscoroutine(task):
                        await task
                    elif iscoroutinefunction(task):
                        await task(*args, **kwargs)
                    else:
                        task(*args, **kwargs)
                except BaseEx as exc:
                    assert False, f'{exc} should be handled within logic'
                except Exception as exc:
                    await error_handler(exc)
                    if self.close_on_exception:
                        self.close(exc)
                        return
        self.close()


@logger_wrapper(name='pymaid.ParallelHandler')
class ParallelHandler(Handler):
    '''Handle the *received* tasks parallelly.

    It holds a worker pool to do the actual work, with a limited concurrency.
    '''

    def __init__(
        self,
        *,
        on_close: Optional[List[Callable[['Handler'], None]]] = None,
        error_handler: Optional[Callable[[Task], Coroutine]] = None,
        close_on_exception: bool = False,
        concurrency: int = 5,
    ):
        super().__init__(
            on_close=on_close,
            error_handler=error_handler,
            close_on_exception=close_on_exception,
        )
        self.worker = AioPool(concurrency)
        self.got_exception = False

    async def _run_callback(self, callback: Callable, *args, **kwargs):
        return callback(*args, **kwargs)

    def _discard_result(self, task: Task):
        if task.cancelled():
            return
        try:
            task.result()
        except BaseEx as exc:
            assert False, f'{exc} should be handled within logic'
            del exc
        except Exception as exc:
            get_running_loop().create_task(self.error_handler(exc))
            if self.close_on_exception:
                self.got_exception = True
                get_running_loop().create_task(self.on_exception(exc))

    async def join(self, reason: Optional[Union[str, Exception]] = None):
        await self.worker.join()
        await super().join(reason)

    async def on_exception(self, exc: Exception):
        await self.worker.join()
        self.close(exc)

    async def run(self):
        new_task_received = self.new_task_received
        pending_tasks = self.pending_tasks
        error_handler = self.error_handler
        schedule_task = self.worker.spawn
        run_callback = self._run_callback

        running = True
        while running:
            await new_task_received.wait()
            # clear for reuse
            new_task_received.clear()
            while pending_tasks:
                task = pending_tasks.popleft()
                if not task:
                    running = False
                    break

                task, args, kwargs = task
                try:
                    if iscoroutine(task):
                        pass
                    elif iscoroutinefunction(task):
                        task = task(*args, **kwargs)
                    else:
                        task = run_callback(task, *args, **kwargs)
                    assert iscoroutine(task), task
                    t = await schedule_task(task, self._discard_result)
                    if self.close_on_exception and self.got_exception:
                        self.logger.warning(
                            'cancel task due to close_on_exception'
                        )
                        t.cancel()
                except Exception as exc:
                    await error_handler(exc)
                    if self.close_on_exception:
                        await self.worker.join()
                        self.close(exc)
                        return
        await self.worker.join()
        self.close()



from io import BytesIO
from queue import deque

import httptools

from multidict import CIMultiDict
from pymaid.conf import settings
from pymaid.net.utils.uri import parse_uri

from .error import HttpError


CRITICAL_ERROR_TEXT = '''HTTP/1.0 500 INTERNAL SERVER ERROR
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Type: text/html; charset=utf-8
Content-Length: 127

<!DOCTYPE HTML PUBLIC '-//W3C//DTD HTML 3.2 Final//EN'>
<title>Internal Server Error</title>
<h1>Internal Server Error</h1>
'''.replace('\n', '\r\n')

HTTP_SWITCHING_PROTOCOLS = '''HTTP/1.1 101 SWITCHING PROTOCOLS
Connection: Upgrade
Upgrade: h2c
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Length: 0

'''.replace('\n', '\r\n')

HTTP_TOO_BIG = '''HTTP/1.1 413 PAYLOAD TOO LARGE
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Length: 0

'''.replace('\n', '\r\n')

HTTP_INVALID_COMPRESSION = '''HTTP/1.1 400 BAD REQUEST
Server: pymaid
X-Powered-By: pymaid
X-HTTP-Backend: httptools
Content-Length: 25

Invalid compressed data
'''.replace('\n', '\r\n')


class Http:

    HEADER_SINGLETON = frozenset(
        [
            'Authorization',
            'Content-Type',
            'Content-Disposition',
            'Content-Length',
            'From',
            'Host',
            'If-Modified_Since',
            'If-Unmodified_Since',
            'Location'
            'Max-Forwards',
            'Proxy-Authorization',
            'Referer',
            'User-Agent',
        ]
    )

    HEADER_MUST_HAVE_VALUE = frozenset(['Location'])

    HEADER_MULTIPLE_VALUE_SPECIAL_CASE = frozenset(
        ['Proxy-Authenticate', 'Set-Cookie', 'WWW-Authenticate']
    )

    def __init__(self):
        self.headers = CIMultiDict()
        self.buffer = BytesIO()

        self.http_version = ''
        self.should_upgrade = False
        self.should_keep_alive = False
        self.chunked = False

    @property
    def body(self):
        return self.buffer.getvalue()

    def append_header(self, name: str, value: str):
        '''Add header.

        When handling multiple values for one header, follow the rule mentioned
        in `Mozilla implementation`_.

        .. _Mozilla implementation: https://github.com/bnoordhuis/mozilla-\
            central/blob/master/netwerk/protocol/http/nsHttpHeaderArray.h#L185
        '''

        # TODO: what to do with `Transfer-Encoding: chunked`
        if name in self.HEADER_SINGLETON and name in self.headers:
            return
            raise HttpError.BadRequest(
                _message_='multiple value for singleton header',
                data={
                    'name': name,
                    'value': value,
                    'in_header': self.headers.getall(name),
                },
            )

        if name in self.HEADER_MUST_HAVE_VALUE and not value:
            raise HttpError.BadRequest(
                _message_='header that must have value get empty value',
                data={'name': name},
            )

        # it seems append to multidict will handle this automatically
        # if name in self.HEADER_MULTIPLE_VALUE_SPECIAL_CASE:
        #     pass

        # assert name and value, f'empty header: name={name} or value={value}'

        # TODO: empty value means no-op
        if not value:
            return
        self.headers.extend([(name, value)])

    # def validate_headers(self):
    #     # called when all headers received
    #     pass

    def extend_body(self, body: bytes):
        self.buffer.write(body)
        if self.buffer.tell() >= settings.pymaid.MAX_BODY_SIZE:
            raise


class HttpRequest(Http):

    def __init__(self):
        super().__init__()
        self.uri = ''
        self.method = ''


class HttpResponse(Http):

    def __init__(self):
        super().__init__()
        self.status = ''
        self.status_code = 0


class Parser:

    '''
    httptools callbacks
    - on_message_begin()
    - on_url(url: bytes)
    - on_header(name: bytes, value: bytes)
    - on_headers_complete()
    - on_body(body: bytes)
    - on_message_complete()
    - on_chunk_header()
    - on_chunk_complete()
    - on_status(status: bytes)

    def get_http_version(self) -> str:
        """Return an HTTP protocol version."""

    def should_keep_alive(self) -> bool:
        """Return ``True`` if keep-alive mode is preferred."""

    def should_upgrade(self) -> bool:
        """Return ``True`` if the parsed request is a valid Upgrade request.
        The method exposes a flag set just before on_headers_complete.
        Calling this method earlier will only yield `False`.
        """

    def get_method(self) -> bytes:
        """Return HTTP request method (GET, HEAD, etc)"""

    def get_status_code(self) -> int:
        """Return the status code of the HTTP response"""

    '''

    def __init__(self):
        self.instance = None
        # TODO: add a max length limitation?
        self.queue = deque()
        self.parser = self.ParserClass(self)

    def feed_data(self, data: bytes) -> int:
        data = memoryview(data)
        # httptools will consume all data
        used = len(data)
        try:
            self.parser.feed_data(data)
        except httptools.HttpParserUpgrade as exc:
            # do nothing about HttpParserUpgrade
            # just return a complete instance for upper level to handle

            # the first args is the offset of used data
            used = exc.args[0]
        except httptools.HttpParserError as exc:
            raise exc
        return used

    def has(self) -> bool:
        return bool(self.queue)

    def get(self) -> Http:
        assert self.queue
        return self.queue.popleft()

    def handle_parser_exception(self, exc):
        '''Default exception handler.

        Raise exc again.
        '''
        raise exc

    def on_message_begin(self):
        '''Called when a message begins. '''
        assert not self.instance
        self.instance = self.ProtocolClass()

    def on_header(self, name: bytes, value: bytes):
        '''Called when a header has been received.

        :param bytes name: The name of the header.
        :param bytes value: The value of the header.
        '''
        self.instance.append_header(name.decode(), value.decode())

    def on_chunk_header(self):
        self.instance.chunked = True

    def on_headers_complete(self):
        '''Called when the headers have been completely sent.'''
        instance = self.instance
        parser = self.parser

        # instance.validate_headers()

        instance.http_version = parser.get_http_version()
        instance.should_upgrade = parser.should_upgrade()
        instance.keep_alive = parser.should_keep_alive()

    def on_body(self, body: bytes):
        '''Called when part of the body has been received.

        :param bytes body: The body bytes.
        '''
        self.instance.extend_body(body)

    def on_message_complete(self):
        '''Put the complete request/response instance into queue.

        Reset instance for the next parse round.
        '''
        self.queue.append(self.instance)
        self.instance = None


class RequestParser(Parser):

    ParserClass = httptools.HttpRequestParser
    ProtocolClass = HttpRequest

    def on_url(self, url: bytes):
        self.instance.uri = parse_uri(url.decode('utf-8'))

    def on_headers_complete(self):
        super().on_headers_complete()
        self.instance.method = self.parser.get_method().decode('utf-8')


class ResponseParser(Parser):

    ParserClass = httptools.HttpResponseParser
    ProtocolClass = HttpResponse

    def on_status(self, status: bytes):
        self.instance.status = status.decode('utf-8')
        self.instance.status_code = self.parser.get_status_code()



from orjson import dumps

from google.protobuf.descriptor_pb2 import MethodDescriptorProto
from google.protobuf.service_reflection import GeneratedServiceType

from pymaid.rpc.method import UnaryUnaryMethod, UnaryStreamMethod
from pymaid.rpc.method import StreamUnaryMethod, StreamStreamMethod
from pymaid.rpc.method import UnaryUnaryMethodStub, UnaryStreamMethodStub
from pymaid.rpc.method import StreamUnaryMethodStub, StreamStreamMethodStub
from pymaid.rpc.router import Router, RouterStub

from .error import PBError
from .pymaid_pb2 import Context as Meta, Void, ErrorMessage


class PBRouter(Router):

    def get_service_methods(self, service: GeneratedServiceType):
        for method in service.DESCRIPTOR.methods:
            method_impl = getattr(service, method.name)
            if '.<lambda>' in str(method_impl):
                continue

            mdp = MethodDescriptorProto()
            method.CopyToProto(mdp)
            if not mdp.client_streaming and not mdp.server_streaming:
                method_class = UnaryUnaryMethod
            elif not mdp.client_streaming and mdp.server_streaming:
                method_class = UnaryStreamMethod
            elif mdp.client_streaming and not mdp.server_streaming:
                method_class = StreamUnaryMethod
            elif mdp.client_streaming and mdp.server_streaming:
                method_class = StreamStreamMethod
            else:
                assert False, 'should be one of above'

            request_class = service.GetRequestClass(method)
            response_class = service.GetResponseClass(method)
            method_ins = method_class(
                method.name,
                method.full_name,
                method_impl,
                request_class,
                response_class,
                options={
                    'flags': Meta.PacketFlag.NULL,
                    'void_request': issubclass(request_class, Void),
                    'void_response': issubclass(response_class, Void),
                },
            )
            yield method_ins

    def feed_messages(self, conn, messages):
        Request = Meta.PacketType.REQUEST
        Response = Meta.PacketType.RESPONSE
        get_route = self.get_route
        tasks = []
        for message in messages:
            # check exist context here for a shortcut
            # because just feed message into context won't block,
            # and it makes serial streaming posible
            meta, payload = message
            # self.logger.debug(f'{self} feed meta={meta}')
            if meta.transmission_id in conn.context_manager.contexts:
                conn.context_manager.contexts[meta.transmission_id] \
                    .feed_message(meta, payload)
                continue

            if meta.packet_type == Request:
                name = meta.service_method
                rpc = get_route(name)
                if rpc is None:
                    task = self.handle_error(
                        meta, PBError.RPCNotFound(data={'name': name})
                    )
                else:
                    context = conn.context_manager.new_inbound_context(
                        meta.transmission_id,
                        method=rpc,
                        conn=conn,
                        timeout=conn.timeout,
                    )
                    context.feed_message(meta, payload)
                    task = context.run()
            elif meta.packet_type == Response:
                # response should be handled above as existed context
                self.logger.warning(
                    f'{self!r} received unknown response, '
                    f'id={meta.transmission_id}, ignored.'
                )
                continue
            else:
                task = self.handle_error(
                    conn,
                    meta,
                    PBError.InvalidPacketType(
                        data={'packet_type': meta.packet_type}
                    )
                )
            tasks.append(task)
        return tasks

    async def handle_error(self, conn, meta, error):
        meta.is_failed = True
        meta.packet_type = Meta.PacketType.RESPONSE
        packet = ErrorMessage(code=error.code, message=error.message)
        if error.data:
            packet.data = dumps(error.data)
        await conn.send_message(meta, packet)


class PBRouterStub(RouterStub):

    def get_router_stubs(self, stub):
        for method in stub.DESCRIPTOR.methods:
            mdp = MethodDescriptorProto()
            method.CopyToProto(mdp)
            if not mdp.client_streaming and not mdp.server_streaming:
                method_class = UnaryUnaryMethodStub
            elif not mdp.client_streaming and mdp.server_streaming:
                method_class = UnaryStreamMethodStub
            elif mdp.client_streaming and not mdp.server_streaming:
                method_class = StreamUnaryMethodStub
            elif mdp.client_streaming and mdp.server_streaming:
                method_class = StreamStreamMethodStub
            else:
                assert False, 'should be one of above'

            request_class = stub.GetRequestClass(method)
            response_class = stub.GetResponseClass(method)
            method_stub = method_class(
                method.name,
                method.full_name,
                request_class,
                response_class,
                options={
                    'flags': Meta.PacketFlag.NULL,
                    'void_request': issubclass(request_class, Void),
                    'void_response': issubclass(response_class, Void),
                },
            )
            yield method_stub



from base64 import b64encode
from io import BytesIO
from os import urandom

from pymaid.core import Event
from pymaid.net.http.h11 import RequestParser, ResponseParser
from pymaid.net.stream import Stream
from pymaid.net.utils.uri import URI
from pymaid.types import DataType
from pymaid.utils.logger import logger_wrapper

from .exceptions import ProtocolError
from .protocol import WSProtocol, Frame, CloseReason
from .utf8validator import Utf8Validator


@logger_wrapper(name='pymaid.WebSocket')
class WebSocket(Stream):
    '''A naive implementation of `websocket`_.

    .. _websocket: https://tools.ietf.org/html/rfc6455
    '''

    PROTOCOL = WSProtocol
    KEEP_OPEN_ON_EOF = True
    REQUIRE_MASK_CLIENT_FRAMES = True

    def __init__(
        self,
        sock,
        *,
        on_open=None,
        on_close=None,
        initiative=False,
        ssl_context=None,
        ssl_handshake_timeout=None,
        uri: URI = None,
        **kwargs,
    ):
        super().__init__(
            sock,
            on_open=on_open,
            on_close=on_close,
            initiative=initiative,
            ssl_context=ssl_context,
            ssl_handshake_timeout=ssl_handshake_timeout,
            uri=uri,
        )
        self.conn_made_event = Event()
        self.resource = ''
        if initiative:
            assert uri is not None
            self.resource = uri.path.encode('utf-8')
            if uri.query:
                self.resource = f'{uri.path}?{uri.query}'.encode('utf-8')
        self.ws_kwargs = kwargs
        self.__read_buffer = BytesIO()
        if self.initiative:
            self._start_handshake()
            self._data_received = self._parse_upgrade_response
            self._handshake_parser = ResponseParser()
        else:
            self._data_received = self._parse_upgrade_request
            self._handshake_parser = RequestParser()
        self.on_close.append(self.cleanup)

        self.current_opcode = None
        self.current_message = []
        self.utf8validator = Utf8Validator()

    async def write(self, message: DataType):
        '''Send a frame over the websocket with message as its payload.'''
        await self._write(
            self.PROTOCOL.encode(
                message, self.get_mask_key() if self.need_mask else b''
            )
        )

    def write_sync(self, message: DataType):
        self._write_sync(
            self.PROTOCOL.encode(
                message, self.get_mask_key() if self.need_mask else b''
            )
        )

    def mark_ready(self):
        # we are finished upgrade handshake
        self.state = self.STATE.CONNECTED
        self.conn_made_event.set()
        self._data_received = self._parse_frames

    def handle_close(self, frame: Frame):
        '''Called when a close frame has been decoded from the stream.

        :param frame: The decoded `Frame`.
        '''
        self._write_sync(
            self.PROTOCOL.encode_frame(Frame.OPCODE_CLOSE, frame.payload),
        )
        self.close(frame.close_reason)

    def handle_ping(self, frame: Frame):
        self._write_sync(
            self.PROTOCOL.encode_frame(Frame.OPCODE_PONG, frame.payload)
        )

    def handle_pong(self, frame: Frame):
        pass

    def get_mask_key(self) -> bytes:
        return urandom(4)

    @staticmethod
    def cleanup(self, exc=None):
        del self.__read_buffer

    @property
    def need_mask(self):
        '''All frames sent from the client to the server are masked by a 32-bit
        value that is contained within the frame.

        https://tools.ietf.org/html/rfc6455#section-5.3

        :NOTE: These rules might be relaxed in a future specification.
        '''
        return self.initiative and self.REQUIRE_MASK_CLIENT_FRAMES

    def _parse_frames(self, data: DataType):
        '''Parse frames from incoming buffer.

        NOTE: correct frames will be lost if some frames are incorrect.
        '''
        buf = self.__read_buffer
        buf.seek(0, 2)
        buf.write(data)
        buf.seek(0, 0)

        mbuf = buf.getbuffer()
        try:
            used_size, frames = self.PROTOCOL.feed_data(mbuf)
        finally:
            mbuf.release()
        if frames:
            buf.seek(used_size, 0)
            self.__read_buffer = BytesIO()
            self.__read_buffer.write(buf.read())
            try:
                self._handle_frames(frames)
            except ProtocolError as ex:
                self._write_sync(
                    self.PROTOCOL.encode_frame(
                        Frame.OPCODE_CLOSE,
                        ex.payload,
                        self.get_mask_key() if self.need_mask else b''
                    )
                )
                self.close(ex.reason)

    def _handle_frames(self, frames) -> bytes:
        opcode = self.current_opcode
        message = self.current_message
        data_received = self.data_received

        for frame in frames:

            if isinstance(frame, ProtocolError):
                raise frame

            if self.initiative and frame.mask:
                raise ProtocolError('masked server-to-client frame')
            elif not self.initiative and not frame.mask:
                raise ProtocolError('unmasked client-to-server frame')

            f_opcode = frame.opcode
            if f_opcode == Frame.OPCODE_TEXT:
                # a new frame
                if opcode:
                    raise ProtocolError(
                        f'The opcode in non-fin frame is expected to be zero, '
                        f'got {f_opcode}'
                    )
                opcode = f_opcode
                # Start reading a new message, reset the validator
                self.utf8validator.reset()
            elif f_opcode == Frame.OPCODE_BINARY:
                # a new frame
                if opcode:
                    raise ProtocolError(
                        f'The opcode in non-fin frame is expected to be zero, '
                        f'got {f_opcode}'
                    )
                opcode = f_opcode
            elif f_opcode == Frame.OPCODE_CONTINUATION:
                if not opcode:
                    raise ProtocolError('Unexpected frame with opcode=0')
            elif f_opcode == Frame.OPCODE_PING:
                self.handle_ping(frame)
                continue
            elif f_opcode == Frame.OPCODE_PONG:
                self.handle_pong(frame)
                continue
            elif f_opcode == Frame.OPCODE_CLOSE:
                self.handle_close(frame)
                return
            else:
                raise ProtocolError(f'Unexpected opcode={f_opcode}')

            if opcode == Frame.OPCODE_TEXT:
                # payload may be empty or one byte
                # making pyaload not a valid code point
                # need to used utf8validator instead
                resp = self.utf8validator.validate(frame.payload)
                if not resp[0]:
                    raise ProtocolError(
                        'Invalid UTF-8 payload',
                        CloseReason.INVALID_FRAME_PAYLOAD_DATA,
                    )
            message.append(frame.payload)

            if frame.fin:
                if opcode == Frame.OPCODE_TEXT:
                    data_received(b''.join(message).decode('utf-8'))
                else:
                    data_received(b''.join(message))
                opcode = None
                del message[:]

        self.current_opcode = opcode

    def _parse_upgrade_request(self, data: DataType) -> bool:
        consumed = self._handshake_parser.feed_data(data)
        if not self._handshake_parser.has():
            # if not has a http instance, parser will consume all data
            return False

        ins = self._handshake_parser.get()
        assert ins.should_upgrade, ins

        resp = self.PROTOCOL.build_response(ins.headers)
        self._write_sync(resp)

        self.mark_ready()
        self.__read_buffer = BytesIO()
        if consumed < len(data):
            self.__read_buffer.write(data[consumed:])
        return True

    def _parse_upgrade_response(self, data: DataType) -> bool:
        consumed = self._handshake_parser.feed_data(data)
        if not self._handshake_parser.has():
            # if not has a http instance, parser will consume all data
            return False

        ins = self._handshake_parser.get()
        assert ins.should_upgrade, ins

        self.PROTOCOL.validate_upgrade(ins.headers, self.secret_key)

        self.mark_ready()
        self.__read_buffer = BytesIO()
        if consumed < len(data):
            self.__read_buffer.write(data[consumed:])
        return True

    def _start_handshake(self):
        self.logger.debug(f'{self.id} start handshake')
        key = self.secret_key = b64encode(urandom(16)).strip()
        self._write_sync(
            self.PROTOCOL.build_request(
                self.uri.host.encode('utf-8'),
                self.resource,
                key,
                **self.ws_kwargs
            )
        )



import ssl as _ssl

from typing import Optional, Tuple, Type, TypeVar, Union

from pymaid.ext.middleware import MiddlewareManager
from pymaid.net.channel import StreamChannel as NetStreamChannel
from pymaid.ext.handler import SerialHandler
from pymaid.net.protocol import Protocol, ProtocolType
from pymaid.types import HandlerType
from pymaid.utils.logger import get_logger

from .connection import Connection, ConnectionType
from .context import ContextManager
from .router import Router

__all__ = ('Channel',)

logger = get_logger('rpc')


class Channel(NetStreamChannel):
    '''Channel manage connections.

    RPC *mostly* rely on ordered data, so it should be in stream mode.
    '''

    logger = logger

    def __init__(
        self,
        *,
        name: str = 'RPCChannel',
        address: Union[Tuple[str, int], str] = '',
        transport_class: ConnectionType = Connection,
        ssl_context: _ssl.SSLContext,
        ssl_handshake_timeout: Optional[float] = None,
        protocol_class: ProtocolType = Protocol,
        handler_class: HandlerType = SerialHandler,
        router_class: Type[Router] = Router,
        context_manager_class: Type[ContextManager] = ContextManager,
        middleware_manager: Optional[MiddlewareManager] = None,
    ):
        super().__init__(
            name=name,
            address=address,
            transport_class=transport_class,
            ssl_context=ssl_context,
            ssl_handshake_timeout=ssl_handshake_timeout,
            middleware_manager=middleware_manager,
        )

        self.protocol_class = protocol_class
        self.handler_class = handler_class
        self.router = router_class()
        self.context_manager_class = context_manager_class

    def _make_connection(
        self,
        sock,
        initiative,
        on_open=None,
        on_close=None,
    ) -> ConnectionType:
        return self.transport_class(
            sock,
            initiative=initiative,
            ssl_context=self.ssl_context,
            ssl_handshake_timeout=self.ssl_handshake_timeout,
            on_open=on_open,
            on_close=on_close,
            protocol=self.protocol_class,
            handler=self.handler_class(),
            router=self.router,
            context_manager=self.context_manager_class(initiative=False),
        )

    def connection_made(self, sock) -> ConnectionType:
        conn = super().connection_made(sock)
        self.middleware_manager.dispatch('on_connection_made', self, conn)
        return conn

    def connection_lost(
        self,
        conn: ConnectionType,
        exc: Optional[Exception] = None
    ):
        # full = self.is_full
        super().connection_lost(conn, exc)
        self.middleware_manager.dispatch('on_connection_lost', self, conn)
        # if (full
        #         and not self.is_full
        #         and not self.is_paused
        #         and not self.is_stopped
        #         and self.middleware_manager.dispatch('can_accept', self)):
        #     self.start()

    def start(self):
        super().start()
        self.middleware_manager.dispatch('on_start', self)

    def pause(self, reason: str = ''):
        super().pause(reason)
        self.middleware_manager.dispatch('on_pause', self)

    def shutdown(self, reason: str = 'shutdown'):
        super().shutdown(reason)
        self.middleware_manager.dispatch('on_shutdown', self)

    def close(
        self, reason: Union[None, str, Exception] = 'called close',
    ):
        super().close(reason)
        self.middleware_manager.dispatch('on_close', self)


ChannelType = TypeVar('Channel', bound=Channel)



'''Raw socket layer.

Mostly inspired from standard lib `asyncio`.
'''

import os
import re
import socket

from errno import ENOTCONN, ECONNABORTED
from typing import List

from pymaid.conf import settings
from pymaid.core import get_running_loop, run_in_threadpool, sleep

HAS_IPv6_FAMILY = hasattr(socket, 'AF_INET6')
HAS_IPv6_PROTOCOL = hasattr(socket, 'IPPROTO_IPV6')

HAS_UNIX_FAMILY = hasattr(socket, 'AF_UNIX')

STREAM_OPTS = {
    'tcp': (socket.AF_UNSPEC, socket.SOCK_STREAM),
    'tcp4': (socket.AF_INET, socket.SOCK_STREAM),
    'tcp6': (socket.AF_INET6, socket.SOCK_STREAM),
    'unix': (socket.AF_UNIX, socket.SOCK_STREAM),
}

DATAGRAM_OPTS = {
    'udp': (socket.AF_UNSPEC, socket.SOCK_DGRAM),
    'udp4': (socket.AF_INET, socket.SOCK_DGRAM),
    'udp6': (socket.AF_INET6, socket.SOCK_DGRAM),
    'unix': (socket.AF_UNIX, socket.SOCK_DGRAM),
}

NET_OPTS = {
    **STREAM_OPTS,
    **DATAGRAM_OPTS,
    'any': (socket.AF_UNSPEC, 0),
    'unix': (socket.AF_UNIX, 0),
}

ADDRESS_REGEX = re.compile(r'([\w\.]+):?(\w*)|\[([\w:]+)\]:?(\w*)')


async def getaddrinfo(
    address: str,
    family: socket.AddressFamily,
    socket_kind: socket.SocketKind,
    flags: int = 0,
):
    if address.startswith('/'):
        if family not in {socket.AF_UNIX, socket.AF_UNSPEC}:
            raise ValueError(
                'address starts with `/` should be unix family, '
                f'got address={address} family={family}'
            )
        if socket_kind == 0:
            infos = [
                (socket.AF_UNIX, socket.SOCK_STREAM, 0, '', address),
                (socket.AF_UNIX, socket.SOCK_DGRAM, 0, '', address),
            ]
        else:
            infos = [(socket.AF_UNIX, socket_kind, 0, '', address)]
    else:
        match = ADDRESS_REGEX.match(address)
        if not match:
            raise ValueError(f'invalid address: {address}')
        host, port = (g for g in match.groups() if g)
        infos = await run_in_threadpool(
            socket.getaddrinfo,
            args=(host, port, family, socket_kind),
            kwargs={'flags': flags},
        )
        infos = list(set(infos))
    return infos


def set_sock_options(sock: socket.socket):
    setsockopt = sock.setsockopt

    # stream opts
    if sock.type == socket.SOCK_STREAM and sock.family != socket.AF_UNIX:
        setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)


async def sock_connect(
    net: str,
    address: str,
    flags: int = 0,
) -> socket.socket:
    loop = get_running_loop()
    err = None
    if net not in STREAM_OPTS:
        raise ValueError(f'only support {STREAM_OPTS.keys()} now, got {net}')
    family, socket_kind = STREAM_OPTS[net]
    addr_infos = await getaddrinfo(address, family, socket_kind, flags)
    for addr_info in addr_infos:
        retry = 3
        af, kind, proto, canonname, sa = addr_info
        sock = None
        while 1:
            try:
                sock = socket.socket(af, kind, proto)
                sock.setblocking(False)
                await loop.sock_connect(sock, sa)
                set_sock_options(sock)
                # NOTE:
                # When doing a lots connect to remote side under heavy pressure
                # it would sometimes getting ENOTCONN when call getpeername.
                # Check it here, if occured, raise it to retry.
                # *WHY* sock_connect above does not handle this case?
                # NOTE 2:
                # This case appears to inconsistently occur with
                # bound to a unix domain socket.
                sock.getpeername()
                # Break explicitly a reference cycle
                err = None
                return sock
            except socket.error as _:
                if _.errno == 107:
                    # OSError: [Errno 107] Transport endpoint is not connected
                    # special case when dealing with 107, it seems retry later
                    # is ok.
                    await sleep(0.001)
                    continue
                if _.errno in {ECONNABORTED, ENOTCONN} and retry:
                    retry -= 1
                    continue
                err = _
                if sock is not None:
                    sock.close()
                break

    if err is not None:
        try:
            raise err
        finally:
            # Break explicitly a reference cycle
            err = None
    else:
        raise socket.error('getaddrinfo returns an empty list')


async def sock_listen(
    net: str,
    address: str,
    flags: socket.AddressInfo = socket.AI_PASSIVE,
    backlog: int = 4096,
    reuse_address: bool = True,
    reuse_port: bool = False,
) -> List[socket.socket]:
    '''Create sockets listening on `address`.

    The address parameter can be a string, in that case the TCP sock is
    bound to unix domain sock.

    The address parameter can also be a tuple of string and int, in that case
    the TCP sock is bound to host and port.
    If a host appears multiple times (possibly indirectly e.g. when hostnames
    resolve to the same IP address), the sock is only bound once to that
    host.

    This is a coroutine.

    :returns: `socket.socket` objects that listening on `address`.
    '''
    if net not in STREAM_OPTS:
        raise ValueError(f'only support {STREAM_OPTS.keys()} now, got {net}')

    sockets = []
    family, socket_kind = STREAM_OPTS[net]
    addr_infos = await getaddrinfo(address, family, socket_kind, flags)
    try:
        for addr_info in addr_infos:
            af, kind, proto, canonname, sa = addr_info
            try:
                sock = socket.socket(af, kind, proto)
            except socket.error:
                # Assume it's a bad family/type/protocol combination.
                continue
            sock.setblocking(False)
            set_sock_options(sock)
            if reuse_address:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
            if settings.get('REUSE_PORT', ns='pymaid') or reuse_port:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            # Disable IPv4/IPv6 dual stack support (enabled by
            # default on Linux) which makes a single socket
            # listen on both address families.
            if HAS_IPv6_FAMILY and HAS_IPv6_PROTOCOL and af == socket.AF_INET6:
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, True)

            if af == socket.AF_UNIX and os.path.exists(sa):
                os.unlink(sa)
            try:
                sock.bind(sa)
            except OSError as err:
                if err.errno == 98:
                    assert af == socket.AF_UNIX, af
                    raise OSError(
                        98,
                        (
                            'Address already in use; '
                            'pymaid cannot bind AF_UNIX parallelly'
                        )
                    ) from None
                raise OSError(
                    err.errno,
                    f'error occured while binding on address {addr_info}: '
                    f'{err.strerror}, addr_infos={addr_infos}'
                ) from None
            # see https://github.com/golang/go/issues/5030
            sock.listen(min(backlog, 65535))
            sockets.append(sock)
    except Exception:
        for sock in sockets:
            sock.close()
        raise

    if not sockets:
        raise socket.error('getaddrinfo returns an empty list')
    return sockets



import argparse

from pymaid import __version__
from pymaid.conf import settings


class ArgumentParser(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop('name', '')
        self.on_parse = kwargs.pop('on_parse', None)
        super().__init__(*args, **kwargs)
        self.subparsers = None

    def create_subparser(
        self,
        name: str,
        **kwargs,
    ):
        group = self.name
        if not self.subparsers:
            self.subparsers = self.add_subparsers(
                title=f'{group} subcommands' if group else 'subcommands',
                dest=f'{group}_subcmd' if group else 'subcmd',
            )

        sub = self.subparsers.add_parser(
            name,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            **kwargs,
        )
        sub.name = f'{group}_{name}' if group else name
        return sub

    def parse_args(self, args=None, namespace=None):
        args = super().parse_args(args, namespace)
        self.on_parse_callback(args)
        return args

    def on_parse_callback(self, args):
        if self.on_parse:
            self.on_parse(args)
        if self.subparsers:
            subcmd = self.get_subcmd(args)
            if subcmd:
                sub_parser = self.subparsers._name_parser_map[subcmd]
                sub_parser.on_parse_callback(args)

    def get_subcmd(self, args):
        subcmd_name = f'{self.name}_subcmd' if self.name else 'subcmd'
        return getattr(args, subcmd_name, None)


def get_parser():
    return parser


def on_parse(args):
    if args.conf:
        settings.load_from_cli(args.conf)
    if args.debug:
        settings.set('DEBUG', True, ns='pymaid')
    if args.loop:
        settings.set('EVENT_LOOP', args.loop, ns='pymaid')
    if args.log_level:
        value = settings.pymaid.LOGGING
        for logger in value['loggers'].values():
            logger['level'] = args.log_level
        settings.set('LOGGING', value, ns='pymaid')


parser = ArgumentParser(
    prog='pymaid',
    description='pymaid, a handy toolset',
    usage='%(prog)s [options]',
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    on_parse=on_parse,
)
parser.add_argument(
    '--version', action='version', version=f'%(prog)s {__version__}',
)
parser.add_argument('--debug', action='store_true', help='enable debug mode')
parser.add_argument(
    '--loop',
    choices=('vanilla', 'uvloop'),
    default='uvloop',
    help='set the event loop policy, default is uvloop for better performance',
)
parser.add_argument(
    '--conf',
    type=str,
    action='append',
    help='set conf, format like NS__KEY=TYPE::VALUE; e.g. USER__AGE=int::18',
)
parser.add_argument(
    '--log-level',
    choices=('FATAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'),
    default=settings.pymaid.LOGGING['loggers']['pymaid']['level'],
    help='set logging level',
)



'''pymaid separate network into two layer

- transport layer
    transport layer concern about how data transmit
    for example: stream for tcp; packet for udp
    and security of transmission: with or without ssl/tls

- protocol layer
    protocol layer concern about how app data format
    it does not care where the data comes from
    stream, packet, unix, or pipeline are not different

    you feed it data, and it gives you protocol messages, that's all

    for example: http, protocol buffer, websocket and so on,
    has the ability to wrap builtin Protocols
    like AppProtocol(PBProtocol) or even AppProtocol(PBProtocol(WSProtocol))
'''

__all__ = ['Protocol', 'Transport', 'Stream', 'Datagram']

import socket
import ssl

from typing import Callable, List, Optional, Union


from .channel import ChannelType, StreamChannel
from .raw import sock_connect
from .stream import Stream
from .utils.uri import parse_uri


async def dial_stream(
    address: str,
    *,
    transport_class: Stream = Stream,
    ssl_context: Union[None, bool, 'ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    on_open: Optional[List[Callable]] = None,
    on_close: Optional[List[Callable]] = None,
    **kwargs,
):
    '''Create a `Stream` instance that connect to `address`.

    The address parameter can be a string, in that case the stream is
    connect to unix domain sock.

    The address parameter can also be a tuple of string and int, in that case
    the stream is connect to the address and port. If a address
    appears multiple times (possibly indirectly e.g. when hostnames
    resolve to the same IP address), the stream is only connect once to that
    address.

    This method is a coroutine.

    :returns: a `Stream` object.
    '''
    uri = parse_uri(address)
    sock = await sock_connect(uri.scheme, uri.address)
    stream = transport_class(
        sock,
        initiative=True,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        on_open=on_open,
        on_close=on_close,
        uri=uri,
        **kwargs,
    )
    await stream.wait_ready()
    return stream


def create_channel(
    channel_class: ChannelType = StreamChannel,
    *,
    transport_class: Stream = Stream,
    ssl_context: Union[None, 'ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    **kwargs,
):
    '''Create channel instance.

    :returns: a `Channel` instance which can be used to manage the streams.
    '''
    return channel_class(
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        **kwargs,
    )


async def serve_stream(
    address: str,
    *,
    name: str = 'StreamChannel',
    channel_class: ChannelType = StreamChannel,
    transport_class: Stream = Stream,
    flags: socket.AddressInfo = socket.AI_PASSIVE,
    backlog: int = 4096,
    reuse_address: bool = True,
    reuse_port: bool = False,
    ssl_context: Union[None, 'ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    start_serving: bool = True,
    **kwargs,
):
    '''Create a channel instance listening on `address` and serve forever.

    The address parameter can be a string, in that case the TCP Channel is
    bound to unix domain sock.

    The address parameter can also be a tuple of string and int, in that case
    the TCP Channel is bound to the address and port.

    If a address appears multiple times (possibly indirectly,
    e.g. when hostnames resolve to the same IP address),
    the Channel is only bound once to that address.

    This method is a coroutine.
    '''
    channel = create_channel(
        channel_class=channel_class,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        **kwargs,
    )
    await channel.listen(
        address,
        flags=flags,
        backlog=backlog,
        reuse_address=reuse_address,
        reuse_port=reuse_port,
    )
    if start_serving:
        channel.start()
    return channel



DEBUG = False
EVENT_LOOP = 'uvloop'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': (
                '%(asctime)s.%(msecs).03d %(levelname).1s|%(name)s: '
                '%(message)s'
            ),
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
            'stream': 'ext://sys.stdout',
        },
    },
    'loggers': {
        'pymaid': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
        'asyncio': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
    },
}

# Sets the maximum number of consecutive accepts that a process may perform
# on a single wake up. High values give higher priority to high connection
# rates, while lower values give higher priority to already established
# connections.
# Default is 64. Note, that in case of multiple working processes on the
# same listening value, it should be set to a lower value.
MAX_ACCEPT = 64
MAX_TASKS = 32

#
# MAX_CONNECTIONS limits the connections
# MAX_CONCURRENCY limits the concurrency of *requests*
# MAX_METHOD_CONCURRENCY limits the concurrency of *requests* for specified rpc
# e.g.:
# 1. 10000 connections, next new connection will fail
# 2. 100 connections, one parallelly call 100 different(not the same) async
# rpcs next rpc call will fail (nomatter from new or old connections)
# 3. 1 connection, parallelly call one async rpc for 10000 times
# next the same rpc call from the connection will fail
#
MAX_CONNECTIONS = 10000
MAX_CONCURRENCY = 10000
MAX_METHOD_CONCURRENCY = 10000

# connection/socket related settings
PM_WEBSOCKET_TIMEOUT = 15
MAX_BODY_SIZE = 10 * 1024 * 1024



import abc

from typing import Any, TypeVar

from pymaid.types import DataType


class Protocol(abc.ABC):
    '''pymaid use Protocol representation app protocol layer

    You can build your protocol upon Protocol
    and you can easily change the underlying Protocol

    Build your AppProtocol inherit from Http and change to Http2 if wanted,
    and in best case, do not need to do anything else

    .. code-block:: python

        class AppProtocol(Http):
            ...

        class AppProtocol(Http2):
            ...

    '''

    @abc.abstractclassmethod
    def feed_data(cls, data: DataType):
        raise NotImplementedError

    @abc.abstractclassmethod
    def encode(cls, obj: Any) -> DataType:
        raise NotImplementedError

    @abc.abstractclassmethod
    def decode(cls, data: DataType) -> Any:
        raise NotImplementedError


ProtocolType = TypeVar('Protocol', bound=Protocol)



import socket
import ssl as _ssl
from typing import Optional, Sequence, Type, Union

from pymaid.ext.handler import SerialHandler
from pymaid.ext.middleware import MiddlewareManager
from pymaid.net.protocol import Protocol, ProtocolType
from pymaid.types import HandlerType

from . import channel
from . import connection
from . import context
from . import pb
from . import router

from .types import ServiceType

__all__ = ('channel', 'connection', 'context', 'pb', 'router')


async def serve_stream(
    address: str,
    *,
    name: str = 'StreamChannel',
    channel_class: channel.ChannelType = channel.Channel,
    transport_class: connection.ConnectionType = connection.Connection,
    family: socket.AddressFamily = socket.AF_UNSPEC,
    flags: socket.AddressInfo = socket.AI_PASSIVE,
    backlog: int = 4096,
    reuse_address: bool = True,
    reuse_port: bool = False,
    ssl_context: Union[None, bool, '_ssl.SSLContext'] = None,
    ssl_handshake_timeout: Optional[float] = None,
    start_serving: bool = True,
    middleware_manager: Optional[MiddlewareManager] = None,
    protocol_class: Type[ProtocolType] = Protocol,
    handler_class: Optional[HandlerType] = SerialHandler,
    router_class: Type[router.Router] = router.Router,
    context_manager_class: context.ContextManager = context.ContextManager,
    services: Optional[Sequence[ServiceType]] = None,
    router: Optional[router.Router] = None,
):
    assert services is not None or router, 'should provide services or router'
    ch = channel_class(
        name=name,
        transport_class=transport_class,
        ssl_context=ssl_context,
        ssl_handshake_timeout=ssl_handshake_timeout,
        middleware_manager=middleware_manager,
        protocol_class=protocol_class,
        handler_class=handler_class,
        router_class=router_class,
        context_manager_class=context_manager_class,
    )
    if services:
        ch.router.include_services(services)
    if router:
        ch.router.include_router(router)
    await ch.listen(
        address,
        family=family,
        flags=flags,
        backlog=backlog,
        reuse_address=reuse_address,
        reuse_port=reuse_port,
    )
    if start_serving:
        ch.start()
    return ch



'''Inspired by `websockets`_.

.. _websockets: https://github.com/aaugustin/websockets/blob/main/src/websockets/uri.py  # noqa
'''
from dataclasses import dataclass
from os.path import isdir
from typing import Optional, Tuple
from urllib.parse import quote, urlsplit

# All characters from the gen-delims and sub-delims sets in RFC 3987.
DELIMS = ":/?#[]@!$&'()*+,;="
SCHEMES = {
    'unix', 'tcp', 'tcp4', 'tcp6', 'udp', 'udp4', 'udp6',
    'http', 'https', 'ws', 'wss',
    '',  # for url from content like `GET /whatever HTTP/1.1`
}


@dataclass
class URI:
    '''URI.

    :param str scheme: scheme
    :param str host: lower-case host
    :param int port: None for `unix` scheme,
        otherwise always set even if it's the default
    :param str path: path, `/` for default
    :param str query: optional query
    :param str fragment: optional query
    :param bool secure: secure flag
    :param str user_info: ``(username, password)`` tuple when the URI contains
        `User Information`_, else ``None``.
    :param str address: a combination of host:port for convenience

    .. _User Information: https://tools.ietf.org/html/rfc3986#section-3.2.1
    '''

    scheme: str
    host: str
    port: Optional[int]
    path: str
    query: str
    fragment: str
    secure: bool
    user_info: Optional[Tuple[str, str]]
    address: str


def parse_uri(uri: str) -> URI:
    '''Parse and validate URI.

    :raises ValueError: if ``uri`` is not a valid URI.
    '''
    parsed = urlsplit(uri)
    scheme = parsed.scheme

    # NOTE:
    #  unix/tcp/udp are layer 4 protocols
    # http/https/ws/wss are layer 7 protocols
    if scheme not in SCHEMES:
        raise ValueError(
            f'only {SCHEMES} scheme support now, '
            f'got uri={uri!r}, scheme={scheme!r}'
        )

    secure = scheme in {'https', 'wss'}
    # pymaid forced ipv6 address format like `[::1]`, sourrounded by `[]`
    host = f'[{parsed.hostname}]' if scheme.endswith('6') else parsed.hostname
    path = parsed.path or '/'
    query = parsed.query
    fragment = parsed.fragment

    if 'unix' == scheme:
        port = None
        # when using unix domain socket, assume path is the address
        if host:
            raise ValueError(
                f'`unix` scheme should not has host, '
                'e.g. unix:///absolute/path/to/sock; '
                f'got uri={uri!r} parsed={parsed!r}'
            )
        if not path or isdir(path):
            raise ValueError(
                '`unix` scheme should be with absolute path, '
                'e.g. unix:///absolute/path/to/sock; '
                f'got uri={uri!r} parsed={parsed!r}'
            )
        # used absolute path as host for unix scheme
        host = path
        path = '/'
    else:
        port = parsed.port or (443 if secure else 80)

    user_info = None
    if parsed.username is not None:
        # urlsplit accepts URLs with a username but without a password.
        # This doesn't make sense for HTTP Basic Auth credentials.
        if parsed.password is None:
            raise ValueError(
                'URI with a username but without a password makes no sense'
            )
        user_info = (parsed.username, parsed.password)

    try:
        uri.encode("ascii")
    except UnicodeEncodeError:
        # Input contains non-ASCII characters.
        # It must be an IRI. Convert it to a URI.
        host = host.encode("idna").decode()
        path = quote(path, safe=DELIMS)
        query = quote(query, safe=DELIMS)
        if user_info is not None:
            user_info = (
                quote(user_info[0], safe=DELIMS),
                quote(user_info[1], safe=DELIMS),
            )

    return URI(
        scheme,
        host,
        port,
        path,
        query,
        fragment,
        secure,
        user_info,
        f'{host}:{port}' if port else host,
    )
